DEBUG = False

CONFIG_FILE_NAME = "config.yaml"
CODE_TAR_GZ = 'code.tar.gz'
DATASET_TAR_GZ = 'dataset.tar.gz'
TAR_GZ_EXTENSION = 'tar.gz'

TEMP_AUTH_TOKEN = 'temp_token'

USER_CODE_EXTENSIONS = ['.py', '.gz']
MODEL_UPLOAD_EXTENSIONS = ['.pyd', '.so', '.build', '.txt', '.yaml']

SUPPORTED_JOB_TYPES = ['tensorflow', 'pytorch']
JOB_TYPE_TO_API = {
    'tensorflow': 'tf',
    'pytorch': 'pt'
}

JOB_CONTENT_TYPE_CODE = "code"
JOB_CONTENT_TYPE_DATASET = "dataset"
JOB_CONTENT_TYPE_OUTPUT = "output"

JOB_TIME_FMT = "%Y-%m-%d %H:%M:%S"

# Active job states
JOB_STATE_PENDING_UPLOAD = "pending upload"
JOB_STATE_UPLOADED = "job data uploaded"
JOB_STATE_READY = "job is ready"

# Inactive job states
JOB_STATE_CANNOT_UPLOAD = "cannot upload job data"
JOB_STATE_CANNOT_SIGNAL_JOB_READY = "failed to signal job ready"
JOB_STATE_SERVER_RESPONSE_UNEXPECTED = "unexpected server response"
JOB_STATE_SERVER_SIGNALED_STOPPED = "server signaled stop"
JOB_STATE_TERMINATED = "user stopped the job"
JOB_STATE_OUTPUT_DOWNLOADED = "output downloaded"
JOB_STATE_NO_OUTPUT = "job has no output"



import os
import subprocess
import traceback
from os.path import join, isdir, exists, normpath
from subprocess import Popen

from clint.textui import prompt, validators

from dccli.constants import *
from dccli.utils import console_out_warn, console_out_log


def pack_code_using_git(out_folder, working_directory):
    # upload current git directory and its sub directories
    repo_tar_path = join(out_folder, CODE_TAR_GZ)

    try:
        # start packaging code into tar.gz
        # check if there is uncommitted change in the working dir
        diff_process = Popen(
            ['git', 'status', '--porcelain'], stdout=subprocess.PIPE)
        diff_process.wait()
        diffs = diff_process.stdout.read().decode("utf-8").split("\n")

        # if there is no diff, process will output empty string
        if len(diffs) > 1:
            # user has uncommitted change in working directory
            console_out_warn("Code package includes uncommitted changes")

            # prompt for untracked files
            check_untracked_process = Popen(['git', 'ls-files', '--others', '--exclude-standard'], stdout=subprocess.PIPE)
            check_untracked_process.wait()
            untracked = diff_process.stdout.read().decode("utf-8").split("\n")
            if len(untracked) > 1:
                add_untracked = prompt.query(
                    "Do you want to add untracked files to training job? [y/n]: ",
                    default=os.getcwd(),
                    validators=[validators.OptionValidator(['y', 'n', 'yes', 'no'])])

                if str(add_untracked).lower() in ['y', 'yes']:
                    add_process = Popen(['git', 'add', '-A'], stdout=subprocess.PIPE)
                    add_process.wait()
                else:
                    console_out_info("skip untracked files", level="info")

            stash_process = Popen(['git', 'stash'], stdout=subprocess.PIPE)
            stash_process.wait()

            archive_process = Popen(
                ['git', 'archive', '--format={}'.format(TAR_GZ_EXTENSION), 'stash@{0}',
                    '--output={}'.format(repo_tar_path)],
                    stderr=subprocess.PIPE,
                    cwd=working_directory)

            console_out_log("zip source code...")
            archive_process.wait()
            pop_stash_process = Popen(
                ['git', 'stash', 'pop'], stdout=subprocess.PIPE)
            pop_stash_process.wait()
        else:
            archive_process = Popen(['git', 'archive', '--format={}'.format(
                TAR_GZ_EXTENSION), 'HEAD', '--output={}'.format(repo_tar_path)],
                stderr=subprocess.PIPE,
                cwd=working_directory)

            console_out_log("zip source code...")
            archive_process.wait()

        if not exists(repo_tar_path):
            return False, "Failed to pack the code"
        else:
            return True, repo_tar_path

    except:
        if DEBUG:
            traceback.print_exc()

        return False, "Failed to pack the code"

def get_repo_toplevel_dir():
    process = Popen(
        ['git', 'rev-parse', '--show-toplevel'], stdout=subprocess.PIPE)
    process.wait()
    toplevel = process.stdout.read().decode("utf-8").strip('\n')
    if not isdir(toplevel):
        return None
    return normpath(toplevel)



from clint import resources

from dccli.git_helper import get_repo_toplevel_dir
from dccli.job import Submit, CheckProgress, Download, StreamLog, Stop
from dccli.user import Login, Register
from dccli.utils import *


def main():
    if get_repo_toplevel_dir() is None:
        console_out_error("Error: Not in a git repository")
        exit(-1)

    support_functions = {
        "register": Register(),
        "login": Login(),
        "submit": Submit(),
        "progress": CheckProgress(),
        "stream": StreamLog(),
        "download": Download(),
        "stop": Stop()
    }

    function_lists = [k for k in support_functions.keys()]

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "function_name",
        type=str,
        help="function name you want to invoke",
        choices=function_lists)

    FLAGS, _ = parser.parse_known_args()
    function_name = FLAGS.function_name
    resources.init('DeepCluster', 'dccli')
    try:
        support_functions[function_name].parse_and_execute()
    except KeyboardInterrupt:
        exit(0)
    except Exception as e:
        console_out_error("Error: Unexpected error {}".format(str(e)))
        if DEBUG:
            traceback.print_exc()
            exit(-1)


if __name__ == "__main__":
    main()



import argparse
import json
import os
import tarfile
import traceback
from datetime import datetime
from os import listdir
from os.path import splitext, basename, join, isfile, isdir, getsize

import requests
import yaml
from clint import resources
from clint.textui import puts, indent, colored, progress, prompt

from dccli.constants import *


def console_out_log(message, level="log", tab=0, quote=''):
    def _puts(level, message):
        if level == "success":
            puts(colored.green(message))
        elif level == "error":
            puts(colored.red(message))
        elif level == "info":
            puts(colored.cyan(message))
        elif level == "warn":
            puts(colored.yellow(message))
        else:
            puts(colored.clean(message))

    # Cannot have quote but no tab.
    if quote != "" and tab == 0:
        tab = 4

    if tab > 0:
        with indent(indent=tab, quote=quote):
            _puts(level, message)
    else:
        _puts(level, message)


def console_out_success(message, tab=0, quote=''):
    console_out_log(message, level="success", tab=tab, quote=quote)


def console_out_error(message, tab=0, quote=''):
    console_out_log(message, level="error", tab=tab, quote=quote)


def console_out_info(message, tab=0, quote=''):
    console_out_log(message, level="info", tab=tab, quote=quote)


def console_out_warn(message, tab=0, quote=''):
    console_out_log(message, level="warn", tab=tab, quote=quote)


class TokenAuth(requests.auth.AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers['Authorization'] = 'Token ' + self.token
        return r


class BaseCase:
    master_endpoint = None
    parser = argparse.ArgumentParser()
    welcome = None
    splitter = "==========================================="

    def __init__(self, master_endpoint="http://dtf-masterserver-dev.us-west-1.elasticbeanstalk.com"):
        self.master_endpoint = master_endpoint

    def add_arguments(self):
        # add case specific arguments
        pass

    def execute(self, FLAGS):
        # add your specific logic functions
        raise Exception("must be overridden")

    def welcome_message(self):
        if self.welcome is not None:
            front = (len(self.splitter) - len(self.welcome)) // 2
            back = (len(self.splitter) - len(self.welcome)) - front
            console_out_log(self.splitter)
            console_out_log("{}{}{}".format(" "*front, self.welcome, " "*back))
            console_out_log(self.splitter)

    def parse_and_execute(self):
        self.add_arguments()
        FLAGS, _ = self.parser.parse_known_args()
        self.welcome_message()
        self.execute(FLAGS)

    def get_job_folder(self, job_uuid):
        return join(resources.user.path, job_uuid)

    def get_job_config_file_name(self, job_uuid):
        return "{}.ini".format(job_uuid)

    def get_job_config(self, job_uuid):
        config = resources.user.read(
            self.get_job_config_file_name(job_uuid))

        if config is None:
            return None

        try:
            config_parsed = json.loads(config)
        except:
            console_out_warn("Failed to load job config for job {}".format(job_uuid))
            if DEBUG:
                traceback.print_exc()

            return None
        return config_parsed

    def update_job_config(self, job_uuid, **kwargs):
        # Attempt to load the config first
        existing_config = self.get_job_config(job_uuid)
        timenow = datetime.now().strftime(JOB_TIME_FMT)
        if existing_config is None:
            existing_config = {
                "create_at": timenow,
                "last_state_at": timenow,
                "job_active": True,
                "output_downloaded": False
            }

        # Load/update configuration.
        for key in kwargs.keys():
            # These keys are auto updated.
            ignore_keys = ["last_state_at", "terminate_at", "create_at", "update_at"]
            if key in ignore_keys:
                continue

            # Note that the entire job state histroy is kept.
            if key == "job_state":
                # Append the new state at the end of the state history.
                if key in existing_config:
                    state_history = existing_config[key]
                else:
                    state_history = []

                state_history += [(kwargs[key], timenow)]
                existing_config[key] = state_history
                existing_config["last_state_at"] = timenow
            else:
                # Record the time that job is terminated.
                if key == "job_active":
                    still_active = existing_config.get("job_active", True)
                    if not kwargs[key] and still_active:
                        existing_config["terminate_at"] = timenow
                existing_config[key] = kwargs[key]

        existing_config["update_at"] = timenow
        resources.user.write(
            self.get_job_config_file_name(job_uuid),
            json.dumps(existing_config))

        return existing_config

    def get_all_job_configs(self):
        job_configs = {}
        user_folder = resources.user.path
        if isdir(user_folder):
            for item in listdir(user_folder):
                item_path = join(user_folder, item)
                if isfile(item_path):
                    filename, extension = splitext(item)
                    if extension == ".ini":
                        job_config = self.get_job_config(filename)
                        if job_config is not None:
                            job_configs[filename] = job_config
        return job_configs

    def select_a_job(self, include_conditions={"job_active": True}, auto_select_only_one=True):
        all_jobs = self.get_all_job_configs()
        options = []
        index = 1
        for job_uuid in all_jobs.keys():
            skip = False
            # If the job does not match condition, exclude it.
            for condition_key in include_conditions.keys():
                if condition_key not in all_jobs[job_uuid]:
                    continue
                if all_jobs[job_uuid][condition_key] != include_conditions[condition_key]:
                    skip = True
                    break
            if skip:
                continue

            options += [{
                'selector': str(index),
                'prompt': "({}) {}: {}".format(
                    all_jobs[job_uuid]["create_at"],
                    all_jobs[job_uuid].get("job_type", "<data lost>"),
                    job_uuid),
                'return': job_uuid
            }]

            index += 1

        if len(options) == 0:
            return None

        if len(options) == 1 and auto_select_only_one:
            return options[0]['return']

        selected = prompt.options(
            "Please select a job (1-{}):".format(len(options)), options)
        return selected

    def display_job_info(self, job_uuid, job_config):
        job_type = job_config.get("job_type", "<data lost>")
        console_out_log("job type:  {}".format(job_type), quote=" >")

        # Job current state from server.
        server_state = job_config.get("server_state", None)
        tensorboard = job_config.get("tensorboard_location", None)
        if server_state is not None:
            console_out_log("job state: {}".format(server_state), quote=" >")

        # Some additional information...
        create_time = datetime.strptime(job_config["create_at"], JOB_TIME_FMT)
        if job_config["job_active"]:
            timenow = datetime.now()
        else:
            timenow = datetime.strptime(
                job_config["terminate_at"], JOB_TIME_FMT)

        duration = (timenow - create_time).total_seconds()
        hours, remainder = divmod(duration, 3600)
        minutes, seconds = divmod(remainder, 60)
        duration_str = '{:02}:{:02}:{:02}'.format(
            int(hours), int(minutes), int(seconds))
        duration_str = '{:02}:{:02}:{:02}'.format(int(hours), int(minutes), int(seconds))
        console_out_log("duration:  {}".format(duration_str), quote=" >")
        if tensorboard is not None:
            console_out_log("tensorboard: {}".format(
                tensorboard), quote=" >")
        job_state = job_config.get("job_state", None)
        if job_state:
            console_out_log("job local history:", quote=" >")
            for (job_state, time) in job_config["job_state"]:
                console_out_log("[{}] {}".format(time, job_state), tab=8)


def load_yaml(task_path):
    with open(task_path) as f:
        config = None
        filename, file_extension = splitext(basename(task_path))
        if file_extension == ".json":
            config = json.load(f)
        elif file_extension == ".yaml":
            config = yaml.load(f)
        else:
            raise ValueError("{} is not a valid config".format(str(f)))
    return config


def get_size(path):
    total_size = 0
    for item in listdir(path):
        sub_path = join(path, item)
        if isfile(sub_path):
            total_size += getsize(sub_path)
        else:
            total_size += get_size(sub_path)
    return total_size


def make_tarfile(output_filename, source_dir):
    tar_info = {
        "total_size": get_size(source_dir),
        "tar_size": 0
    }

    def _progress_filter(tarinfo):
        total_size = tar_info["total_size"]
        tar_size = tar_info["tar_size"]
        file_size = tarinfo.size
        tar_size += file_size
        current_progress = (tar_size * 100) // total_size
        tar_info["tar_size"] = tar_size
        bar.show(current_progress)
        return tarinfo

    if source_dir is not None:
        if not source_dir.endswith(os.path.sep):
            source_dir = source_dir + os.path.sep
        with progress.Bar(label="zip dataset ", expected_size=100) as bar:
            bar.show(0)
            with tarfile.open(output_filename, "w:gz") as tar:
                tar.add(source_dir, arcname=basename(source_dir), filter=_progress_filter)


def unzip_tarfile(source_tar, target_dir):
    tar = tarfile.open(source_tar, "r:gz")
    tar.extractall(path=target_dir)
    tar.close()


def rreplace(s, old, new, occurrence):
    li = s.rsplit(old, occurrence)
    return new.join(li)


def request_server(url, verb, **kwargs):
    response = None
    # Allow a 30s timeout for the request.
    try:
        if verb == 'post':
            response = requests.post(url, timeout=30, **kwargs)
        elif verb == 'put':
            response = requests.put(url, timeout=30, **kwargs)
        elif verb == 'get':
            response = requests.get(url, timeout=30, **kwargs)
        else:
            console_out_error("Unknown api '{}'".format(verb))
            return False, None, None, None
    except Exception as e:
        console_out_error("Server is temporarily unavailable...")
        if DEBUG:
            traceback.print_exc()
        return False, None, None, None

    if response is None:
        console_out_error("Server is temporarily unavailable...")
        return False, None, None, None

    response_data = None
    try:
        if response.text is not None:
            response_data = json.loads(response.text)
    except Exception as e:
        console_out_error("Server returned malformed data...")
        if DEBUG:
            traceback.print_exc()
        return False, None, None, None

    error = None
    if not response.ok:
        error = "Unknown error..."
        if response_data is not None:
            error = response_data.get("error", "Unknown error...")
            if type(error) is dict:
                error_string = ""
                for k in error.keys():
                    if error[k] is None:
                        continue
                    error_entry = error[k]
                    if type(error[k]) is list:
                        error_entry = error[k][0]
                    if error_string == "":
                        error_string = error_entry
                    else:
                        error_string += (", " + error_entry)
                error = error_string

    return response.ok, response, response_data, error



import getpass
from clint import resources
from clint.textui import prompt, validators
import re

from dccli.error_codes import *
from dccli.utils import *


def get_login_token():
    token = resources.user.read(TEMP_AUTH_TOKEN)
    if token is None:
        return None

    return TokenAuth(token)


class Login(BaseCase):
    welcome = "Login to DeepCluster"

    def execute(self, FLAGS):
        email = prompt.query("Please enter your email:")
        password = getpass.getpass("Please enter your password:")

        data = {
            'email': email,
            'password': password
        }

        return self.login(data)

    def login(self, data):
        server_url = "{}/api/v1/user/login/".format(self.master_endpoint)
        result, response, response_data, error = request_server(server_url, "post", data=data)
        if not result:
            if not response:
                console_out_error("Error: {}".format(error))
                return MASTER_FAILURE
            elif response.status_code == 401:
                console_out_error("Error: Unknown email/password combination")
                return LOGIN_NEED_REGISTER
            else:
                console_out_error("Error: {}".format(error))
                return MASTER_FAILURE
        elif response_data is None or 'auth_token' not in response_data:
            console_out_error("Error: Malformed response from server")
            return MASTER_FAILURE

        auth_token = response_data.get('auth_token')
        resources.user.write(TEMP_AUTH_TOKEN, auth_token)
        console_out_success("Login successfully")
        return SUCCESS


class Register(BaseCase):
    welcome = "Register with DeepCluster"

    def execute(self, FLAGS):
        email = prompt.query("Please enter your email address:")
        console_out_info(
            "Password must be 8 - 20 characters and can contain alphanumeric and @#$%^&+=")
        password = getpass.getpass("Please enter your password: ")
        password_re = re.compile('^[A-Za-z0-9@#$%^&+=]{8,20}$')
        if not password_re.match(password):
            console_out_error("Error: Password invalid")
            return USER_INPUT_ERROR

        re_password = getpass.getpass("Please re-enter your password: ")

        if password != re_password:
            console_out_error("Error: Passwords do not match")
            return USER_INPUT_ERROR

        data = {
            'email': email,
            'password': password
        }

        return self.register(data)

    def register(self, data):
        server_url = "{}/api/v1/user/register/".format(self.master_endpoint)
        result, response, _, error = request_server(server_url, "post", data=data)
        if not result:
            if not response:
                console_out_error("Error: {}".format(error))
                return MASTER_FAILURE
            else:
                console_out_error("Error: {}".format(error))
                return MASTER_FAILURE
        else:
            console_out_success("Registered successfully")
            log_in_case = Login(self.master_endpoint)
            log_in_case.login(data)
            return SUCCESS



from os.path import join, dirname, exists

from dccli.error_codes import SUCCESS
from dccli.job import Submit, Download, CheckProgress, StreamLog
from dccli.user import Register, Login
from test.test_common import test_config, clean_up_login_token, s3_client
from test.test_user import test_login_user

_master_endpoint = test_config["master_endpoint"]
_output_path = join(dirname(__file__))

register = Register(_master_endpoint)
user_data = {
    'email': test_config["email"],
    'password': test_config["password"]
}
register.register(user_data)

# start a new login
clean_up_login_token()
login = Login(_master_endpoint)
login.login(user_data)

_job_uuid = None
_test_maker = join(dirname(__file__), 'testmarker')


def test_end2end_submit():
    test_login_user()

    class DummyFlag:
        c = True

    submit_case = None
    try:
        submit_case = Submit(_master_endpoint)
        assert submit_case.execute(DummyFlag()) == SUCCESS
    finally:
        print(submit_case.job_uuid)
        _job_uuid = submit_case.job_uuid
        with open(_test_maker, 'w') as f:
            f.write(_job_uuid)

        if submit_case is not None and _job_uuid is not None:
            s3_client.s3_delete_by_key(bucket=test_config["s3_bucket_name"], s3_folder=_job_uuid, key=None)


def test_download_output():
    if not exists(_test_maker):
        test_end2end_submit()

    with open(_test_maker, 'r') as f:
        _job_uuid = f.read()

    class DummyFlag:
        job_uuid = _job_uuid
        dest = _output_path

    download_case = Download(_master_endpoint)
    print(DummyFlag.job_uuid)
    assert download_case.execute(DummyFlag) == SUCCESS


def test_check_progress():
    if not exists(_test_maker):
        test_end2end_submit()

    with open(_test_maker, 'r') as f:
        _job_uuid = f.read()

    class DummyFlag:
        job_uuid = _job_uuid

    progress_case = CheckProgress(_master_endpoint)
    assert progress_case.execute(DummyFlag) == SUCCESS


def test_stream_log():
    if not exists(_test_maker):
        test_end2end_submit()

    with open(_test_maker, 'r') as f:
        _job_uuid = f.read()

    class DummyFlag:
        job_uuid = _job_uuid

    log_case = StreamLog(_master_endpoint)
    assert log_case.execute(DummyFlag) == SUCCESS



import sys
from os.path import dirname, join, exists
import yaml
from six import StringIO
from ddls3utils.s3utils import S3Client
import os

from dccli.constants import TEMP_AUTH_TOKEN, USER_CODE_EXTENSIONS, MODEL_UPLOAD_EXTENSIONS

test_config_path = join(dirname(__file__), "test_config.yaml")
with open(test_config_path) as f:
    test_config = yaml.load(f)

s3_client = S3Client(region_name="us-west-2",
                     aws_access_key_id="AKIAJOHKUPVKMZ66GTSQ",
                     aws_secret_access_key="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",
                     download_file_extension_filter=USER_CODE_EXTENSIONS,
                     upload_file_extension_filter=MODEL_UPLOAD_EXTENSIONS)



def capture_console_log(func):
    def wrapper(console_out):
        if "TEST" in os.environ:
            old_stdout = sys.stdout
            sys.stdout = std_out = StringIO()
            console_out = std_out.getvalue()
            func(console_out)
            sys.stdout = old_stdout
            print(console_out)
        else:
            func(None)
    return wrapper


def clean_up_login_token():
    temp_token_path = join(dirname(__file__), TEMP_AUTH_TOKEN)
    if exists(temp_token_path):
        os.remove(temp_token_path)




import uuid
from os.path import join, dirname

from dccli.error_codes import SUCCESS
from dccli.user import Register, Login
from test.test_common import test_config
from clint import resources

master_endpoint = test_config["master_endpoint"]
job_uuid = test_config["job_uuid"]
output_path = join(dirname(__file__), "test_out")

resources.init('DeepCluster', 'dccli')


def test_register_user():
    register_case = Register(master_endpoint)

    rand_email = f"{str(uuid.uuid4())[0:8]}@uw.edu"
    user_data = {
        'email': rand_email,
        'password': test_config["password"]
    }

    assert register_case.register(user_data) == SUCCESS


def test_login_user():
    login = Login(master_endpoint)
    user_data = {
        'email': test_config["email"],
        'password': test_config["password"]
    }

    assert login.login(user_data) == SUCCESS



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="dccli",
    version="0.0.10",
    author="YL & SW",
    author_email="deepcluster.io@gmail.com",
    description="cli for manage job in DeepCluster.io",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(exclude=['test', ]),
    entry_points={
        'console_scripts': [
            'dccli=dccli.cli:main',
        ],
    },
    install_requires=[
        'pyyaml',
        'requests',
        'clint'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import shutil
from time import sleep

from clint.textui import validators

from dccli.error_codes import *
from dccli.git_helper import *
from dccli.user import get_login_token
from dccli.utils import *


class Download(BaseCase):
    welcome = "Download Job Output"

    def add_arguments(self):
        self.parser.add_argument(
            '--dest',
            type=str,
            default=None,
            help="Destination folder to save the outputs")

        self.parser.add_argument(
            '--job_uuid',
            type=str,
            default=None,
            help="Your job uuid shown when you submit a training job")

    def execute(self, FLAGS):
        auth = get_login_token()
        if auth is None:
            console_out_error("Error: You need to login first")
            return USER_NEED_LOGIN

        job_uuid = FLAGS.job_uuid
        if job_uuid is None:
            job_uuid = self.select_a_job(
                include_conditions={"output_downloaded": False})

        if job_uuid is None:
            console_out_warn("No job specified")
            return USER_INPUT_ERROR

        download_to = FLAGS.dest
        if download_to is not None and isfile(download_to):
            console_out_error("Destination is a file")
            return USER_INPUT_ERROR

        if download_to is None or not exists(download_to):
            download_to = prompt.query(
                "Please specify location for download:",
                default=os.getcwd(),
                validators=[validators.PathValidator()])

        return self.download(auth, job_uuid, download_to)

    def download(self, auth, job_uuid, download_to):
        console_out_info("Download For Job {}".format(job_uuid))
        # Only download if the job is finished.
        progress_case = CheckProgress(self.master_endpoint)
        progress_case.check_progress(auth, job_uuid, quiet=True)

        job_config = self.get_job_config(job_uuid)
        if job_config is not None and job_config.get("job_active", False):
            console_out_warn("Download is available only when job finishes")
            return SUCCESS

        server_url = "{}/api/v1/user/download/{}/".format(
            self.master_endpoint, job_uuid)

        result, response, response_data, error = request_server(
            server_url, "get", auth=auth)

        # Validate server response.
        if not result:
            if response is not None and response.status_code == 401:
                console_out_error("Error: You need to login first")
                return USER_NEED_LOGIN

            console_out_error("Error: {}".format(error))
            return MASTER_FAILURE

        if response_data is None:
            console_out_error(
                "Error: Malformed response from server, no data reponsed")
            return MASTER_FAILURE

        download_urls = response_data.get("download_urls", None)
        if download_urls is None:
            console_out_warn("No downloadable content found")
            return MASTER_FAILURE

        output = download_urls.get(JOB_CONTENT_TYPE_OUTPUT, None)
        if output is None:
            console_out_warn("No downloadable content found")
            return MASTER_FAILURE

        # Start downloading.
        downloadables = []
        total_size = 0
        for (hint, url) in output:
            response = requests.get(url, allow_redirects=True)
            if not response.ok:
                continue
            size = int(response.headers.get('content-length', 0))
            if size != 0:
                total_size += size
            else:
                total_size += 1000
            downloadables += [(response, hint, size)]

        if len(downloadables) == 0:
            self.update_job_config(
                job_uuid,
                job_state=JOB_STATE_NO_OUTPUT,
                output_downloaded=True)

            console_out_warn("No downloadable content found")
            return INTERNAL_ERROR

        with progress.Bar(label="download output ", expected_size=100) as bar:
            download_size = 0
            bar.show(0)
            for (downloadable, hint, size) in downloadables:
                file_loc = join(download_to, hint)
                with open(file_loc, 'wb') as fd:
                    for chunk in downloadable.iter_content(chunk_size=128):
                        if size != 0:
                            download_size += 128
                            current_progress = (download_size * 100) // total_size
                            bar.show(current_progress)
                        fd.write(chunk)

                    if size == 0:
                        download_size += 1000
                        current_progress = (download_size * 100) // total_size
                        bar.show(current_progress)

        self.update_job_config(
            job_uuid,
            job_state=JOB_STATE_OUTPUT_DOWNLOADED,
            output_downloaded=True,
            output_folder=download_to)

        console_out_success("Output downloaded to: {}".format(download_to))
        return SUCCESS


class CheckProgress(BaseCase):
    welcome = "Check Job Progress"

    def add_arguments(self):
        self.parser.add_argument(
            '--job_uuid',
            type=str,
            default=None,
            help="Your job uuid shown when you submit the job")

    def execute(self, FLAGS):
        auth = get_login_token()
        if auth is None:
            console_out_error("Error: You need to login first")
            return USER_NEED_LOGIN

        job_uuid = FLAGS.job_uuid
        if job_uuid is None:
            job_uuid = self.select_a_job()

        if job_uuid is None:
            console_out_warn("No job specified")
            return USER_INPUT_ERROR

        return self.check_progress(auth, job_uuid)

    def check_progress(self, auth, job_uuid, quiet=False):
        job_config = self.get_job_config(job_uuid)
        if job_config is not None and not job_config.get("job_active", True):
            # The job is already terminated. Short cut.
            if not quiet:
                console_out_info("Job {}".format(job_uuid))
                self.display_job_info(job_uuid, job_config)
            return SUCCESS

        server_url = "{}/api/v1/user/progress/{}/".format(
            self.master_endpoint, job_uuid)

        result, response, response_data, error = request_server(
            server_url, "get", auth=auth)

        # Validate server response.
        if not result:
            if response is not None and response.status_code == 401:
                if not quiet:
                    console_out_error("Error: You need to login first")
                return USER_NEED_LOGIN

            console_out_error("Error: {}".format(error))
            return MASTER_FAILURE

        if response_data is None:
            if not quiet:
                console_out_error(
                    "Error: Malformed response from server, no data reponsed")
            return MASTER_FAILURE

        # Extract data from server response.
        job_uuid = response_data.get("job_uuid", None)
        if job_uuid is None:
            if not quiet:
                console_out_error(
                    "Error: Malformed response from server, missing 'job_uuid'")
            return MASTER_FAILURE

        job_state = response_data.get("job_state", None)
        if job_state is None:
            if not quiet:
                console_out_error(
                    "Error: Malformed response from server, missing 'job_state'")
            return MASTER_FAILURE

        job_active = response_data.get("job_active", None)
        if job_state is None:
            if not quiet:
                console_out_error(
                    "Error: Malformed response from server, missing 'job_active'")
            return MASTER_FAILURE

        tensorboard_location = response_data.get("tensorboard_location", None)

        # Everything looks right, keep track the updated information.
        job_config_state_update = {
            "server_state": job_state,
            "job_active": job_active
        }

        if not job_active:
            job_config_state_update["job_state"] = JOB_STATE_SERVER_SIGNALED_STOPPED

        if tensorboard_location is not None:
            job_config_state_update["tensorboard_location"] = tensorboard_location

        job_config = self.update_job_config(job_uuid, **job_config_state_update)
        if not quiet:
            console_out_info("Job {}".format(job_uuid))
            self.display_job_info(job_uuid, job_config)

        return SUCCESS


class StreamLog(BaseCase):
    welcome = "Stream Job Logs"

    def add_arguments(self):
        self.parser.add_argument(
            '--job_uuid',
            type=str,
            default=None,
            help="Your job uuid shown when you submit the job")

    def execute(self, FLAGS):
        auth = get_login_token()
        if auth is None:
            console_out_error("Error: You need to login first")
            return USER_NEED_LOGIN

        job_uuid = FLAGS.job_uuid
        if job_uuid is None:
            job_uuid = self.select_a_job()

        if job_uuid is None:
            console_out_warn("No job specified")
            return USER_INPUT_ERROR

        return self.stream_log(auth, job_uuid)

    def stream_log(self, auth, job_uuid):
        has_all_log = False
        last_timestamp = None
        while not has_all_log:
            time_stamp_argument = ""
            if last_timestamp is not None:
                time_stamp_argument = "&timestamp={}".format(last_timestamp)

            server_url = "{}/api/v1/user/stream/?job_uuid={}{}".format(
                self.master_endpoint,
                job_uuid,
                time_stamp_argument)

            result, response, response_data, error = request_server(
                server_url, "get", auth=auth)

            # Validate server response.
            if not result:
                if response is not None and response.status_code == 401:
                    console_out_error("Error: You need to login first")
                    return USER_NEED_LOGIN

                console_out_error("Error: {}".format(error))
                return MASTER_FAILURE

            if response_data is None:
                console_out_error(
                    "Error: Malformed response from server, no data reponsed")
                return MASTER_FAILURE

            # 200 means that is the last batch of log,
            # 202 means there is more
            status_code = response.status_code
            if status_code == 200:
                has_all_log = True

            logs = json.loads(response_data.get('log', None))
            last_timestamp = response_data.get('last_timestamp', None)
            if logs is not None and len(logs) > 0:
                for log in logs:
                    console_out_log(log.rstrip('\n'))
            sleep(3)

        return SUCCESS


class Submit(BaseCase):
    use_current_directory = False
    welcome = "Submit Job"
    job_uuid = None

    def add_arguments(self):
        self.parser.add_argument(
            '-c',
            action='store_true',
            default=False,
            help="Submit the current directory only (instead of the entire repo)")

    def execute(self, FLAGS):
        self.use_current_directory = FLAGS.c
        auth = get_login_token()
        if auth is None:
            console_out_error("Error: You need to login first")
            return USER_NEED_LOGIN

        config_path = self.get_config_file_path()
        if config_path is None or not exists(config_path):
            console_out_warn(
                "No job config found (Expected: '{}')".format(config_path))
            job_data = self.get_job_data_interactive()
            if job_data is None:
                return USER_INPUT_ERROR
        else:
            try:
                job_data = load_yaml(config_path)
            except:
                console_out_error(
                    "Error: cannot load job config at '{}'".format(config_path))
                return USER_INPUT_ERROR

        return self.submit_job(job_data, auth)

    def get_job_toplevel_dir(self):
        if self.use_current_directory:
            toplevel = os.getcwd()
        else:
            toplevel = get_repo_toplevel_dir()
        return toplevel

    def get_config_file_path(self):
        toplevel = self.get_job_toplevel_dir()
        if toplevel is None:
            return None

        return join(toplevel, CONFIG_FILE_NAME)

    def get_job_data_interactive(self):
        job_type = prompt.query(
            "Please specify job type {}:".format(SUPPORTED_JOB_TYPES))

        if job_type.lower() not in SUPPORTED_JOB_TYPES:
            console_out_error(
                "Error: Unsupported job type '{}'".format(job_type))
            return None

        job_data = {
            "job_type": job_type.lower(),
            "worker_required": 1
        }

        yes = prompt.yn("Use local dataset")
        if yes:
            dataset_path = prompt.query(
                "Please specify local dataset location:", validators=[validators.PathValidator()])
            job_data["dataset_path"] = dataset_path
        else:
            dataset_name = prompt.query(
                "Please specify dataset name:", default="cifar10")
            job_data["dataset_name"] = dataset_name

        return job_data

    def validate_job_param(self, job_data):
        job_type = job_data.get("job_type", None)
        if job_type is None:
            raise ValueError(
                "Error: Malformed job configration, missing 'job_type'")

        dataset_name = job_data.get("dataset_name", None)
        dataset_path = job_data.get("dataset_path", None)
        if dataset_path is not None:
            if not isdir(dataset_path) and not isfile(dataset_path):
                raise ValueError(
                    "Error: Malformed job configration, cannot find dataset path '{}'".format(
                        dataset_path))

        return job_type, dataset_name, dataset_path

    def submit_job(self, job_data, auth):
        try:
            job_type, dataset_name, dataset_path = self.validate_job_param(
                job_data)
        except Exception as e:
            console_out_error(str(e))
            return USER_INPUT_ERROR

        request_body = {
            'parameters': json.dumps(job_data)
        }

        # If the dataset name is supplied, assume this is a known dataset and pass that to server.
        if dataset_name is not None:
            request_body['dataset_name'] = dataset_name

        # Submit the job to server to get PreSigned URLs for uploading the code and datasets.
        server_url = "{}/api/v1/user/{}/job/".format(
            self.master_endpoint,
            JOB_TYPE_TO_API[job_type])

        result, response, response_data, error = request_server(
            server_url, "post", data=request_body, auth=auth)

        # Validate server response.
        if not result:
            if response is not None and response.status_code == 401:
                console_out_error("Error: You need to login first")
                return USER_NEED_LOGIN

            console_out_error("Error: {}".format(error))
            return MASTER_FAILURE

        if response_data is None:
            console_out_error(
                "Error: Malformed response from server, no data reponsed")
            return MASTER_FAILURE

        job_uuid = response_data.get("job_uuid", None)
        if job_uuid is None:
            console_out_error(
                "Error: Malformed response from server, missing 'job_uuid'")
            return MASTER_FAILURE

        # Create a config file for the job locally now that we have a job_uuid.
        self.update_job_config(
            job_uuid,
            job_type=job_type,
            source=self.get_job_toplevel_dir(),
            dataset=dataset_name if dataset_name is not None else dataset_path,
            job_state=JOB_STATE_PENDING_UPLOAD)

        # Stop the job if error occurs.
        stop_case = Stop(self.master_endpoint)
        dataset_known = response_data.get("dataset_known", False)
        upload_urls = response_data.get("upload_urls", None)
        if upload_urls is None:
            console_out_error(
                "Error: Malformed response from server, missing 'upload_urls'")

            stop_case.terminate_job(
                job_uuid, auth, JOB_STATE_SERVER_RESPONSE_UNEXPECTED, quiet=True)
            return MASTER_FAILURE

        # Code must be in the upload url.
        upload_code_url = upload_urls.get(JOB_CONTENT_TYPE_CODE, None)
        if upload_code_url is None:
            console_out_error("Error: Malformed response from server, missing '{}'".format(
                JOB_CONTENT_TYPE_CODE))

            stop_case.terminate_job(
                job_uuid, auth, JOB_STATE_SERVER_RESPONSE_UNEXPECTED, quiet=True)
            return MASTER_FAILURE

        # If dataset is not known, dataset must be in the upload url.
        upload_dataset_url = upload_urls.get(
            JOB_CONTENT_TYPE_DATASET, None)

        # Sanity check, we must have a dataset_path if there is upload_dataset_url.
        if dataset_path is not None and upload_dataset_url is None:
            console_out_error("Error: Server does not allow uploading dataset for current config")
            stop_case.terminate_job(
                job_uuid, auth, JOB_STATE_SERVER_RESPONSE_UNEXPECTED, quiet=True)
            return INTERNAL_ERROR

        # Everything is as expected now. Upload the code (and dataset if needed) to the upload url.
        success = self.upload_job_data(
            job_uuid, upload_code_url, upload_dataset_url, dataset_path)

        if not success:
            stop_case.terminate_job(
                job_uuid, auth, JOB_STATE_CANNOT_UPLOAD, quiet=True)
            return INTERNAL_ERROR

        # Job data uploaded.
        self.update_job_config(
            job_uuid,
            job_state=JOB_STATE_UPLOADED)

        # Signal the server that the job is ready.
        signal_server_url = "{}{}/".format(server_url, job_uuid)
        result, response, _, error = request_server(
            signal_server_url, "put", auth=auth)
        if not result:
            console_out_error(
                "Error: Failed to signal job ready, {}".format(error))

            stop_case.terminate_job(
                job_uuid, auth, JOB_STATE_CANNOT_SIGNAL_JOB_READY, quiet=True)
            return MASTER_FAILURE

        console_out_success("Submit job successfully")
        console_out_log("job type: {}".format(job_type), quote=" >")
        console_out_log("job uuid: {}".format(job_uuid), quote=" >")
        # Update the job state to submitted.
        self.update_job_config(
            job_uuid,
            job_state=JOB_STATE_READY)

        # save job_uuid in Submit case so that test can remove the s3 folder
        self.job_uuid = job_uuid
        return SUCCESS

    def pack_dataset(self, out_folder, dataset_path):
        dataset_tar_location = join(out_folder, DATASET_TAR_GZ)
        make_tarfile(dataset_tar_location, dataset_path)
        return dataset_tar_location

    def upload_job_data(self, job_uuid, upload_code_url, upload_dataset_url, dataset_path):
        code_tar_location = None
        dataset_tar_location = None
        job_tmp_folder = self.get_job_folder(job_uuid)

        try:
            # Make a job folder to ensure there is no collision.
            os.mkdir(job_tmp_folder)
            if not isdir(job_tmp_folder):
                console_out_error("Error: Cannot make a job folder")
                return False

            # Pack the code using git.
            success, out = pack_code_using_git(
                job_tmp_folder, self.get_job_toplevel_dir())
            if not success:
                console_out_error("Error: {}".format(out))
                return False

            # out is the path to the code tar ball.
            code_tar_location = out
            console_out_log("Upload code...")
            with open(code_tar_location, 'rb') as fd:
                files = {'file': (CODE_TAR_GZ, fd)}
                response = requests.post(
                    upload_code_url["url"],
                    data=upload_code_url["fields"],
                    files=files)

                if not response.ok:
                    print("Error: Failed to upload code to {}".format(
                        upload_code_url["url"]))
                    return False

            # If we also need to upload dataset, pack the dataset.
            if upload_dataset_url is not None:
                dataset_tar_location = self.pack_dataset(
                    job_tmp_folder, dataset_path)

                console_out_log("Upload dataset...")
                with open(dataset_tar_location, 'rb') as fd:
                    files = {'file': (DATASET_TAR_GZ, fd)}
                    response = requests.post(
                        upload_dataset_url["url"],
                        data=upload_dataset_url["fields"],
                        files=files)

                    if not response.ok:
                        console_out_error("Error: Failed to upload dataset to {}".format(
                            upload_dataset_url["url"]))
                        return False

            return True
        except:
            console_out_error("Error: Failed to upload job data")
            if DEBUG:
                traceback.print_exc()

            return False

        finally:
            if isdir(job_tmp_folder):
                shutil.rmtree(job_tmp_folder, ignore_errors=True)


class Stop(BaseCase):
    welcome = "Stop Job"

    def add_arguments(self):
        self.parser.add_argument(
            '--job_uuid',
            type=str,
            default=None,
            help="Your job uuid shown when you submit the job")

    def execute(self, FLAGS):
        auth = get_login_token()
        if auth is None:
            console_out_error("Error: You need to login first")
            return USER_NEED_LOGIN

        job_uuid = FLAGS.job_uuid
        if job_uuid is None:
            job_uuid = self.select_a_job()

        if job_uuid is None:
            console_out_warn("No job specified")
            return USER_INPUT_ERROR

        # If the job is already stopped, do not make redundant call.
        job_config = self.get_job_config(job_uuid)
        if job_config is not None and not job_config["job_active"]:
            console_out_success("Job {} already stopped".format(job_uuid))
            self.display_job_info(job_uuid, job_config)
            return SUCCESS

        return self.terminate_job(job_uuid, auth, JOB_STATE_TERMINATED)

    def terminate_job(self, job_uuid, auth, final_local_state, quiet=False):
        server_url = "{}/api/v1/user/progress/".format(self.master_endpoint)
        request_body = {
            'job_uuid': job_uuid,
            'job_update_state': "terminate"
        }

        # Tell the server to stop the job.
        result, response, _, error = request_server(
            server_url, "post", data=request_body, auth=auth)

        if not result:
            if response is not None and response.status_code == 401:
                if not quiet:
                    console_out_error("Error: You need to login first")
                return USER_NEED_LOGIN

            if not quiet:
                console_out_error("Error: {}".format(error))
            return MASTER_FAILURE

        # Mark the job as stopped locally.
        job_config = self.update_job_config(
            job_uuid, job_state=final_local_state, job_active=False)
        if not quiet:
            console_out_success("Job {} stopped".format(job_uuid))
            # Some additional information...
            self.display_job_info(job_uuid, job_config)

        return SUCCESS
